/**
 * @file
 * test.adjustColorBrightness node implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"
#include "VuoGlContext.h"

#include <OpenGL/CGLMacro.h>

VuoModuleMetadata({
                     "title" : "Custom Heightfield Shader",
                     "description" : "Stylized heightfield effect that can extrude in any direction. 2014, George Toledo.",

                     "keywords" : [ "paint", "draw", "opengl", "glsl", "scenegraph", "graphics",
                     "outline", "edges", "lines", "skeletal"  ],
                     "version" : "1.0.0",
                     "dependencies" : [
                         "VuoGlContext"
                     ],
                     "node": {
                         "exampleCompositions" : [ "?" ]
                     }
                 });
////////////////////////////
static const char *vertexShaderSource = VUOSHADER_GLSL_SOURCE(120,
    // Inputs provided by VuoSceneRenderer
    uniform mat4 projectionMatrix;
    uniform mat4 modelviewMatrix;
    uniform sampler2D texture;
    uniform float amountX;
    uniform float amountY;
    uniform float amountZ;

    attribute vec4 position;
    attribute vec4 textureCoordinate;

    // Outputs to geometry shader
    varying vec2 textureCoordinateForGeometry;

    void main()
    {
        vec4 tx=texture2D(texture,textureCoordinate.xy);

        gl_Position = projectionMatrix * modelviewMatrix * (position + vec4(length(tx.rgb) * amountX, length(tx.rgb) * amountY, length(tx.rgb) * amountZ,0.));
        textureCoordinateForGeometry = textureCoordinate.xy;
    }
);
/////////////////////////////

static const char *geometryShaderSource = VUOSHADER_GLSL_SOURCE(120,
    // Inputs from vertex shader
    varying in vec2 textureCoordinateForGeometry[3];

    // Outputs to fragment shader
    varying out vec3 geometryDistanceFromEdge;
    varying out vec2 geometryTextureCoordinate;

    void main()
    {
        gl_Position = gl_PositionIn[0];
        geometryDistanceFromEdge = vec3(1,0,0);
        geometryTextureCoordinate = textureCoordinateForGeometry[0];
        EmitVertex();

        gl_Position = gl_PositionIn[1];
        geometryDistanceFromEdge = vec3(0,1,0);
        geometryTextureCoordinate = textureCoordinateForGeometry[1];
        EmitVertex();

        gl_Position = gl_PositionIn[2];
        geometryDistanceFromEdge = vec3(0,0,1);
        geometryTextureCoordinate = textureCoordinateForGeometry[2];
        EmitVertex();

        EndPrimitive();
    }
);

    static const char *pointGeometryShaderSource = VUOSHADER_GLSL_SOURCE(120,
        // Inputs from VuoSceneRenderer
        uniform float aspectRatio;
        uniform float primitiveHalfSize;

        // Inputs from vertex shader
        varying in vec2 textureCoordinateForGeometry[1];

        // Outputs to fragment shader
        varying out vec3 geometryDistanceFromEdge;
        varying out vec2 geometryTextureCoordinate;

        void main()
        {
            vec2 pointSize = vec2(primitiveHalfSize, primitiveHalfSize * aspectRatio);

            gl_Position               = gl_PositionIn[0]       + vec4(-pointSize.x,  pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(1,0,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       + vec4(-pointSize.x, -pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(0,1,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       + vec4( pointSize.x, -pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(0,0,1);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            EndPrimitive();

            gl_Position               = gl_PositionIn[0]       + vec4( pointSize.x, -pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(1,0,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       + vec4( pointSize.x,  pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(0,1,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       + vec4(-pointSize.x,  pointSize.y, 0, 0);
            geometryDistanceFromEdge  = vec3(0,0,1);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            EndPrimitive();
        }
    );

    static const char *lineGeometryShaderSource = VUOSHADER_GLSL_SOURCE(120,
        // Inputs from VuoSceneRenderer
        uniform mat4 projectionMatrix;
        uniform float aspectRatio;
        uniform float primitiveHalfSize;
        uniform vec3 cameraPosition;

        // Inputs from vertex shader
        varying in vec2 textureCoordinateForGeometry[2];

        // Outputs to fragment shader
        varying out vec3 geometryDistanceFromEdge;
        varying out vec2 geometryTextureCoordinate;

        void main()
        {
            vec2 lineSize = vec2(primitiveHalfSize, primitiveHalfSize * aspectRatio);
            vec3 perpendicular = normalize(cross(gl_PositionIn[1].xyz - gl_PositionIn[0].xyz, gl_PositionIn[0].xyz - (projectionMatrix * vec4(cameraPosition,1)).xyz));
            vec4 perpendicularOffset = vec4(perpendicular.x*lineSize.x, perpendicular.y*lineSize.y, 0, 0);

            gl_Position               = gl_PositionIn[1]       - perpendicularOffset;
            geometryDistanceFromEdge  = vec3(1,0,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[1];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       - perpendicularOffset;
            geometryDistanceFromEdge  = vec3(0,1,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[0]       + perpendicularOffset;
            geometryDistanceFromEdge  = vec3(0,0,1);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            EndPrimitive();

            gl_Position               = gl_PositionIn[0]       + perpendicularOffset;
            geometryDistanceFromEdge  = vec3(1,0,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[0];
            EmitVertex();
            gl_Position               = gl_PositionIn[1]       + perpendicularOffset;
            geometryDistanceFromEdge  = vec3(0,1,0);
            geometryTextureCoordinate = textureCoordinateForGeometry[1];
            EmitVertex();
            gl_Position               = gl_PositionIn[1]       - perpendicularOffset;
            geometryDistanceFromEdge  = vec3(0,0,1);
            geometryTextureCoordinate = textureCoordinateForGeometry[1];
            EmitVertex();
            EndPrimitive();
        }
    );
////////////////////
static const char *fragmentShaderSource = VUOSHADER_GLSL_SOURCE(120,
    // Inputs from ports
    uniform vec4 frontLineColor;
    uniform vec4 backLineColor;
    uniform float width;
    uniform float standoffWidth;
    uniform bool showThirdEdge;
    uniform sampler2D texture;
    // Inputs from geometry shader
    varying vec3 geometryDistanceFromEdge;
    varying vec2 geometryTextureCoordinate;
    //varying vec2 textureCoordinateForGeometry;

    void main()
    {
        vec4 tx=texture2D(texture,geometryTextureCoordinate);

        //

        vec2 filterWidth = fwidth(geometryTextureCoordinate);
        //		float maximumFilterWidth = max(filterWidth.x, filterWidth.y) * 8;

                float minimumDistanceFromEdge = min(geometryDistanceFromEdge.x, geometryDistanceFromEdge.z);
                if (showThirdEdge)
                    minimumDistanceFromEdge = min(minimumDistanceFromEdge, geometryDistanceFromEdge.y);

                if (minimumDistanceFromEdge > width+standoffWidth)
                    discard;

        //		float interiorness = smoothstep(width-maximumFilterWidth, width+maximumFilterWidth, minimumDistanceFromEdge);
        //		gl_FragColor = mix(gl_FrontFacing?frontLineColor:backLineColor, vec4(0,0,0,1), interiorness);
                gl_FragColor = minimumDistanceFromEdge < width ? (gl_FrontFacing?frontLineColor:backLineColor) : vec4(0,0,0,1);
                gl_FragColor *= tx;//*vec4(geometryTextureCoordinate.x,geometryTextureCoordinate.y,1.,1.);
    }
);

void nodeEvent
(
        VuoInputData(VuoReal, {"default":0., "suggestedStep":0.05}) amountX,
        VuoInputData(VuoReal, {"default":0., "suggestedStep":0.05}) amountY,
        VuoInputData(VuoReal, {"default":1., "suggestedStep":0.05}) amountZ,
        VuoInputData(VuoColor, {"default":{"r":1.0,"g":1.0,"b":1.0,"a":1.0}}) frontColor,
        VuoInputData(VuoColor, {"default":{"r":1.0,"g":1.0,"b":1.0,"a":1.0}}) backColor,
        VuoInputData(VuoReal, {"default":1.0, "suggestedMin":0.0, "suggestedMax":1.0, "suggestedStep":0.05}) width,
        VuoInputData(VuoReal, {"default":0.1, "suggestedMin":0.0, "suggestedMax":1.0, "suggestedStep":0.05}) standoffWidth,
        VuoInputData(VuoBoolean, {"default":true}) showThirdEdge,
        VuoInputData(VuoImage) image,
        VuoOutputData(VuoShader) shader
)
{
    *shader = VuoShader_make("Custom Heightfield Shader");

    VuoShader_setUniform_VuoImage(*shader, "texture", image);
    VuoShader_setUniform_VuoReal(*shader, "amountX", amountX);
    VuoShader_setUniform_VuoReal(*shader, "amountY", amountY);
    VuoShader_setUniform_VuoReal(*shader, "amountZ", amountZ);


    VuoShader_addSource                      (*shader, VuoMesh_Points,              vertexShaderSource, pointGeometryShaderSource, fragmentShaderSource);
    VuoShader_setExpectedOutputPrimitiveCount(*shader, VuoMesh_Points, 2);

    VuoShader_addSource                      (*shader, VuoMesh_IndividualLines,     vertexShaderSource, lineGeometryShaderSource,  fragmentShaderSource);
    VuoShader_setExpectedOutputPrimitiveCount(*shader, VuoMesh_IndividualLines, 2);

    VuoShader_addSource                      (*shader, VuoMesh_IndividualTriangles, vertexShaderSource, geometryShaderSource,      fragmentShaderSource);

    VuoShader_setUniform_VuoColor(*shader, "frontLineColor", frontColor);
    VuoShader_setUniform_VuoColor(*shader, "backLineColor", backColor);
    VuoShader_setUniform_VuoReal(*shader, "width", width/2.);
    VuoShader_setUniform_VuoReal(*shader, "standoffWidth", standoffWidth/2.);
    VuoShader_setUniform_VuoBoolean(*shader, "showThirdEdge", showThirdEdge);

    {
        CGLContextObj cgl_ctx = (CGLContextObj)VuoGlContext_use();

        // Ensure the command queue gets executed before we return,
        // since the VuoShader might immediately be used on another context.
        glFlushRenderAPPLE();

        VuoGlContext_disuse(cgl_ctx);
    }
}
